from .base_chat import AbstractEntryPoint
from .entrypoints import (
    AiriChatEntryPoint,
    FusionBrainEntrypoint,
    GigaChatCensoredEntryPoint,
    GigaChatEntryPoint,
    GigaMax2EntryPoint,
    GigaMax2SberdevicesEntryPoint,
    GigaMaxEntryPoint,
    GigaPlusEntryPoint,
    OpenRouterEntryPoint,
    YandexGPTEntryPoint,
)
from .entrypoints_accessor import EntrypointsAccessor, create_entrypoint
from .entrypoints_config import EntrypointConfig, EntrypointsConfig

__all__ = [
    "AbstractEntryPoint",
    "OpenRouterEntryPoint",
    "AiriChatEntryPoint",
    "YandexGPTEntryPoint",
    "GigaChatCensoredEntryPoint",
    "GigaChatEntryPoint",
    "GigaPlusEntryPoint",
    "GigaMaxEntryPoint",
    "GigaMax2EntryPoint",
    "GigaMax2SberdevicesEntryPoint",
    "FusionBrainEntrypoint",
    "create_entrypoint",
    "EntrypointsAccessor",
    "EntrypointsConfig",
    "EntrypointConfig",
]



import logging
import time
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from itertools import chain
from typing import Any

import tiktoken
from httpx import NetworkError
from openai.types.chat import ChatCompletionMessageParam
from mmar_llm.models import EntrypointPayload

logger = logging.getLogger(__name__)


class AbstractEntryPoint(ABC):
    @abstractmethod
    def __call__(self) -> Any:
        pass

    @abstractmethod
    def get_response(self, sentence: str) -> str:
        pass

    @abstractmethod
    def get_response_by_payload(self, payload: EntrypointPayload) -> str:
        pass

    @abstractmethod
    def get_embedding(self, sentence: str) -> list[float]:
        pass

    @abstractmethod
    def get_embeddings(self, sentences: list[str], request_limit: int = 50) -> list[list[float]]:
        pass

    def decorate_with(self, decorator) -> "AbstractEntryPoint":
        return EntryPointFacade(self, decorator)

    def get_response_with_retries(self, prompt: str, retries: int = 3) -> str:
        response = self.get_response(prompt)
        retry_count = 0
        if response == "":
            while retry_count < retries:
                time.sleep(5)
                logger.warning(f"Response is empty, attempt {retry_count}")
                retry_count += 1
                response = self.get_response(prompt)
                if response != "":
                    break
        if response == "":
            logger.error("Response is empty!")
        if retry_count > 0:
            logger.info(f"Non-empty answer after retry (count={retry_count})")
        return response

    def get_responses(self, sentences: list[str], max_workers: int = 2) -> list[str]:
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = [executor.submit(self.get_response, sentence) for sentence in sentences]
            responses = [future.result() for future in futures]
        return responses

    def get_responses_by_payload(
        self, payloads: list[list[ChatCompletionMessageParam]], max_workers: int = 2
    ) -> list[str]:
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = [executor.submit(self.get_response_by_payload, payload) for payload in payloads]
            responses = [future.result() for future in futures]
        return responses

    def get_more_embeddings(self, sentences: list[str], batch_size: int = 2, max_workers: int = 4) -> list[list[float]]:
        batches: list[list[str]] = self.make_batches(sentences, size=batch_size)
        if max_workers == 1:
            emb_batches = [self.get_embeddings(batch) for batch in batches]
        else:
            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                futures = [executor.submit(self.get_embeddings, batch) for batch in batches]
                emb_batches = [future.result() for future in futures]
        return list(chain.from_iterable(emb_batches))

    @staticmethod
    def count_tokens(sentences: list[str]) -> list[int]:
        encoding = tiktoken.get_encoding("cl100k_base")
        return [len(encoding.encode(sentence)) for sentence in sentences]

    @staticmethod
    def make_batches(items: list, size: int = 500) -> list[list[str]]:
        slices = [(i * size, (i + 1) * size) for i in range(len(items) // size + 1)]
        return [items[st:ed] for st, ed in slices]

    def warmup(self) -> None:
        response = self.get_response("Прогрев")
        if not response or not sum(self.get_embedding("Прогрев")):
            raise NetworkError("Нет доступа к ллм!")

    @staticmethod
    def create_payload(system_prompt: str, user_prompt: str) -> list[ChatCompletionMessageParam]:
        return [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": user_prompt},
        ]

    @staticmethod
    def create_image_payload(system_prompt: str, user_prompt: str, image_encoded: str, mimetype="image/jpeg"):
        image_content = [
            {"type": "text", "text": user_prompt},
            {"type": "image_url", "image_url": {"url": f"data:{mimetype};base64,{image_encoded}"}},
        ]
        return [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": image_content},
        ]


class EntryPointFacade(AbstractEntryPoint):
    def __init__(self, base: AbstractEntryPoint, decorator):
        self._base = base

        # with basic class (not ABC) it's possible to override method directly
        # like this: `self.get_response = decorator(self._base.get_response)`
        self._get_response = decorator(self._base.get_response)
        self._get_response_by_payload = decorator(self._base.get_response_by_payload)
        self._get_embedding = decorator(self._base.get_embedding)
        self._get_embeddings = decorator(self._base.get_embeddings)

    def __call__(self) -> Any:
        return self._base.__call__()

    def get_response(self, sentence: str) -> str:
        return self._get_response(sentence)

    def get_response_by_payload(self, payload: list[ChatCompletionMessageParam]) -> str:
        return self._get_response_by_payload(payload)

    def get_embedding(self, sentence: str) -> list[float]:
        return self._get_embedding(sentence)

    def get_embeddings(self, sentences: list[str], request_limit: int = 50) -> list[list[float]]:
        return self._get_embeddings(sentences, request_limit=request_limit)



import logging
import re
import warnings
from base64 import b64decode, b64encode
from functools import partial
from time import sleep
from typing import Literal

from gigachat import GigaChat
from gigachat._types import FileTypes
from gigachat.models import Embedding, UploadedFile

from .base_chat import AbstractEntryPoint
from .io_sugar import OnError, make_on_exception
from .models import EntrypointPayload

logger = logging.getLogger(__name__)
PATTERN_CID =  "[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"
CPATTERN_CID = re.compile(PATTERN_CID)


def _validate_cid(field_value, field: str):
    if CPATTERN_CID.fullmatch(field_value):
        return
    warnings.warn(f"Maybe '{field}' is invalid: not matched by '{PATTERN_CID}'")


_validate_client_id = partial(_validate_cid, field="client_id")
_validate_client_secret = partial(_validate_cid, field="client_secret")


def _make_kwargs_auth(
    user: str | None,
    password: str | None,
    client_id: str | None,
    client_secret: str | None,
    authorization_key: str | None,
):
    if user and password:
        return dict(user=user, password=password)
    if client_id and client_secret:
        _validate_client_id(client_id)
        _validate_client_secret(client_secret)
        creds = b64encode(f"{client_id}:{client_secret}".encode("utf-8")).decode("utf-8")
        return dict(credentials=creds)
    if authorization_key:
        authorization_key_decode = b64decode(authorization_key).decode()
        if ":" not in authorization_key_decode:
            warnings.warn("Maybe bad value for 'authorization_key': expected `:`")
        else:
            client_id, client_secret = authorization_key_decode.split(":", 1)
            _validate_client_id(client_id)
            _validate_client_secret(client_secret)
        return dict(credentials=authorization_key)
    raise ValueError(f"Unexpected combination of auth args: {user=}, {password=}, {client_id=}, {client_secret=}")


BASE_URL_DEFAULT = "https://gigachat.devices.sberbank.ru/api/v1"


class GigaChatEntryPoint(AbstractEntryPoint):
    def __init__(
        self,
        user: str | None = None,
        password: str | None = None,
        client_id: str | None = None,
        client_secret: str | None = None,
        authorization_key: str | None = None,
        model_id: str = "GigaChat-Pro",
        warmup: bool = False,
        temperature: float = 0.0,
        base_url: str = BASE_URL_DEFAULT,
        profanity_check: bool = False,
        on_error: OnError = "warn",
        timeout: float | None = None,
    ) -> None:
        kwargs_pre = dict(
            base_url=base_url,
            scope="GIGACHAT_API_CORP",
            model=model_id,
            verify_ssl_certs=False,
            profanity_check=False,
            timeout=timeout,
        )
        kwargs_auth = _make_kwargs_auth(user, password, client_id, client_secret, authorization_key)
        kwargs = kwargs_pre | kwargs_auth
        self._base_url = base_url
        self._model_id = model_id
        self._model = GigaChat(**kwargs)
        self._DIM: int = 1024
        self._ZEROS: list[float] = [0.0 for _ in range(self._DIM)]
        self._ERROR_MESSAGE: str = ""
        self.temperature = temperature
        if warmup:
            self.warmup()
        self.on_error = on_error
        self.on_exception = make_on_exception(on_error, logger, err_msg=f"No response from {model_id}")

    def __call__(self) -> GigaChat:
        return self._model

    def get_response(self, sentence: str) -> str:
        with self.on_exception():
            return self._model.chat(sentence).choices[0].message.content
        return self._ERROR_MESSAGE

    def get_response_by_payload(self, payload: EntrypointPayload) -> str:
        """
        payload:
        - [{"role": "system", "content": system}, {"role": "user", "content": replica}]
        - {"messages": [{"role": "system", "content": system}, {"role": "user", "content": replica}]}
        - {"messages": [...], "attachments": ...}
        """
        if isinstance(payload, list):
            payload_giga = {"messages": payload}
        else:
            payload_giga = payload
        payload_giga["temperature"] = self.temperature
        with self.on_exception():
            return self._model.chat(payload_giga).choices[0].message.content
        return self._ERROR_MESSAGE

    def get_embedding(self, sentence: str) -> list[float]:
        with self.on_exception():
            return self._model.embeddings([sentence]).data[0].embedding
        return self._ZEROS

    def get_embeddings(self, sentences: list[str], request_limit: int = 50) -> list[list[float]]:
        embeddings: list[list[float]] | None = None
        counter: int = 0
        while embeddings is None and counter < request_limit:
            with self.on_exception():
                items: list[Embedding] = self._model.embeddings(sentences).data
                embeddings = [item.embedding for item in items]
                break
            sleep(0.1)
            counter += 1
        if embeddings is not None:
            return embeddings
        return [self._ZEROS for _ in sentences]

    def upload_file(
        self,
        file: FileTypes,
        purpose: Literal["general", "assistant"] = "general",
    ) -> UploadedFile:
        """
        Usage example:
        with Path(fpath).open('rb') as file:
            uploaded_file = entrypoint_giga.upload_file(file=file)
        """
        return self._model.upload_file(file, purpose)

    def __repr__(self):
        class_name = type(self).__name__
        url = self._base_url
        url_info = "" if BASE_URL_DEFAULT == url else f", url={url}"
        model_info = f"model_id={self._model_id}"
        return f"{class_name}({model_info}{url_info})"

    def __str__(self):
        return repr(self)


def _G(**kwargs_base):
    def instantiate_gigachat(**kwargs):
        return GigaChatEntryPoint(**kwargs_base, **kwargs)

    return instantiate_gigachat


GigaChatCensoredEntryPoint = _G(model_id="GigaChat-Pro", profanity_check=True)
GigaMaxEntryPoint = _G(model_id="GigaChat-Max")
GigaMax2EntryPoint = _G(model_id="GigaChat-2-Max")
GigaPlusEntryPoint = _G(model_id="GigaChat-Plus")
GigaMax2SberdevicesEntryPoint = _G(model_id="GigaChat-2-Max", base_url="https://gigachat.sberdevices.ru/v1")
